%let progname = "ebv014.sas";
* xref:
* input:        ebvsim.sas  ci.sas
* output:
* does: Compute the exact distribution from various distribs
        with the same mean and var
********************************************************************;
%let m    = 10;
%let pi   = 0.3;
%let rho  = 0.3;
%let seed = 13579;      * to init the generator;
title1 " Extra-binomial variation, various models  ";
********************************************************************;
proc iml
     worksize=30000  symsize=256; * <- adjust if necessary;
********************************************************************;
%include "ebvsim.sas";
********************************************************************;
start cdf_pdf  (cdf, pdf, dist, n, pi, rho);
/*
cdf, pdf <- values
*/

  if (upcase(dist)="MAD") then
    run madsen_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="BB") then
    run betabin_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="MN") then
    run morelnag_d (cdf, pdf, n, pi, rho);
  else if (upcase(dist)="BIN") then
    run binomial_d (cdf, pdf, n, pi     );
  else do;
    print "cdf_pdf: Unrecognized distribution" dist;
    cdf=.; pdf=.;
    return;
  end;

finish;
********************************************************************;
/* iml execution starts here -> */

* get macro variables;
m  = &m;
pi = &pi;
rho= &rho;

distribs = {
  "BIN"
  "BB"
  "MN"
  "MAD"
}`;

all_pdf = j(m+1, nrow(distribs), .);

do distrib = 1 to nrow(distribs);

  distrib_name = distribs[distrib];
  run   cdf_pdf (cdf, pdf, distrib_name, m, pi, rho);
* print  distrib_name  pi rho cdf pdf ;
  all_pdf[,distrib] = pdf;

end;

name = distribs`;
print  name ;
print  all_pdf ;
********************************************************************;

