* ebvsim.sas
* xref: Zou & Donner (2004) Biometrics
        Madsen (1993) Comm. Stat. A
        Morel & Nagaraj (1993) Bka
        Morel & Neerchal(1997) Stat. Med.
* does: Simulate extra-binomial variation, various models
        IML include file
********************************************************************;

start  get_freq (value, freq, y);
/*
    value  <- values
    freq <- freqs
    y = input data, col or row, integers only
*/
  a = min(y);
  b = max(y);
  value= (a:b)`;
  freq = j(b-a+1, 1, 0);
  n = nrow(y)*ncol(y);
  do i = 1 to n;                   * unavoidable loop ???;
    j = y[i] - a + 1;
    freq[j] = freq[j] + 1;
  end;

finish;

********************************************************************;
start  moments   (mom, y, f);
/*
    mom[1:4] <- moments
    skappa[1:4] <- stndzd cumulants
    y = values, col vec
    f = pdf or freqs, col vec
*/
  mom  = j(4, 1, .);
  p = f`/sum(f);        * row vec;
  mom[1] = p * y;
  mom[2] = p * y##2;
  mom[3] = p * y##3;
  mom[4] = p * y##4;

finish;

********************************************************************;
start  cumulants (kappa, skappa, y, f);
/*
    kappa[1:4] <- cumulants
    skappa[1:4] <- stndzd cumulants
    y = values, col vec
    f = pdf or freqs, col vec
*/
  kappa  = j(4, 1, .);
  skappa = j(4, 1, .);
  p = f`/sum(f);        * row vec;
  kappa[1] = p * y;
  r = y-kappa[1];
  kappa[2] = p * r##2;
  kappa[3] = p * r##3;
  kappa[4] = p * r##4 - 3 # kappa[2]##2;
  skappa[1] = kappa[1] / sqrt(kappa[2]);
  skappa[2] = 1;
  skappa[3] = kappa[3] / kappa[2]##1.5;
  skappa[4] = kappa[4] / kappa[2]##2;

finish;

********************************************************************;
start binomial_d(cdf, pdf, m, pi);
/*
    cdf  <- binomial(m, pi) cdf
    pdf  <- binomial(m, pi) pdf
    m, pi = parameters, scalars
*/


  i=(0:m)`;

  cdf = probbnml (pi, m, i);
  pdf = cdf - (0 // cdf[1:m]);
finish;
********************************************************************;
start betabin_d(cdf, pdf, m, pi, rho);
/*
    cdf  <- betabin(m, pi, rho) cdf
    pdf  <- betabin(m, pi, rho) pdf
    m, pi, rho = parameters, scalars
    rho = 0 gives the binomial cdf and pdf.
*/

  if (rho = 0) then do;
    run binomial_d(cdf, pdf, m, pi);
    return;
  end;

  t = (1-rho)/rho;
  v = t*pi;
  w = t*(1-pi);

  y=(0:m)`;
  pdf = lgamma(m+1)-lgamma(y+1)-lgamma(m-y+1)
      + lbeta(v+y,w+m-y)-lbeta(v,w);
  pdf = exp(pdf);
  pdf = pdf/sum(pdf);
  cdf = cusum(pdf);
  cdf[m+1] = 1;

finish;
********************************************************************;
start lbeta(x,y);
/*
    log Beta(x, y)
*/
  return ( lgamma(x) + lgamma(y) - lgamma(x+y) );

finish;
********************************************************************;
start threept_d(cdf, pdf, m, pi, rho);
/*
    cdf, pdf <- 3-point distribution cdf and pdf (3 x 1 vecs)
    point mass at y=0, m*pi, m
    m, pi, rho = parameters, scalars
*/

  phi_m  = (1+rho*(m-1)) / m;   * phi / m;
  pdf    = j(3, 1, 1-phi_m);    * pdf[2] = 1-phi/m = the mass at pi*m;
  pdf[1] = (1-pi) * phi_m;      * mass at 0;
  pdf[3] =    pi  * phi_m;      * mass at m;
  cdf    = pdf;
  cdf[2] = pdf[1] + pdf[2];
  cdf[3] = 1;

finish;
********************************************************************;
start morelnag_d (cdf, pdf, m, pi, rho);
/*
    cdf  <- Morel-Nagaraaj(m, pi, rho) cdf
    pdf  <- Morel-Nagaraaj(m, pi, rho) pdf
    m, pi, rho = parameters, scalars
*/


  alpha = sqrt(rho)#(1-pi) + pi;
  beta  = pi#(1-sqrt(rho));
  run binomial_d(cdf1, pdf1, m, alpha);
  run binomial_d(cdf0, pdf0, m, beta);

  pdf = pi * pdf1 + (1-pi) * pdf0;
  cdf = pi * cdf1 + (1-pi) * cdf0;

finish;

********************************************************************;
start madsen_d (cdf, pdf, m, pi, rho);
/*
    cdf  <- madsen(m, pi, rho) cdf
    pdf  <- madsen(m, pi, rho) pdf
    m, pi, rho = parameters, scalars
*/
  pdf1 = j(m+1, 1, 0);
  pdf1[1]   = 1 - pi;
  pdf1[m+1] = pi;                      * m Bern(pi) wp rho;
  run binomial_d(cdf0, pdf0, m, pi);   * Bin(m,pi)  wp 1-rho;

  pdf = rho * pdf1 + (1-rho) * pdf0;
  cdf = cusum(pdf);
finish;

********************************************************************;
start binomial_v (K, m, pi, rho);
/*
  Returns a col vector of K random variates from the binomial
  K  = number of variates required, integer > 0
  m  = index, integer > 0,   scalar or vector
  pi = parameter, in (0,1),  scalar or vector
  rho= (ignored)
  If any of m, pi and rho is a vector, it must be column vector, K by 1
*/

  if (nrow(m) = 1 & nrow(pi) = 1) then
    seed = j(K, 1, &seed);
  else
    seed = &seed;
  y = ranbin(seed, m, pi);
  return (y);

finish;

********************************************************************;
start betabin_v (K, m, pi, rho);
/*
  Returns a col vector of K random variates from the beta-binomial type mixture
  K  = number of variates required, integer > 0
  m  = index, integer > 0,   scalar or vector
  pi = parameter, in (0,1),  scalar or vector
  rho= parameter, in [0,1),  scalar or vector
  If any of m, pi and rho is a vector, it must be column vector, K by 1
*/

  if (rho = 0) then return(binomial_v(K, m, pi, 0));

  c = (1-rho) / rho;
  a = pi # c;
  b = (1 - pi) # c;
  if (nrow(a) = 1) then
     do; a = j(K, 1, a); b = j(K, 1, b); end;
  p = rand("beta", a, b);
  if (&sysver >= 9) then
    y = rand("binomial",  p, m);        * sas 9;
  else
    y = rand("binomial",  m, p);        * sas 8;

  return (y);
finish;

********************************************************************;
start morelnag_1 (m, pi, rho);
/*
  returns a random variate from the Morel-Nagaraaj type mixture
  m  = index (integer scalar)
  pi = parameter, scalar in (0,1)
  rho= parameter, scalar in (0,1)
*/
  if (ranuni(&seed) <= pi) then
     y  = ranbin(&seed, m, sqrt(rho)*(1-pi) + pi);
  else
      y = ranbin(&seed, m, pi*(1-sqrt(rho)));
  return (y);
finish;

********************************************************************;
start morelnag_v (K, m, pi, rho);
/*
  Returns a col vec of K random variates from the Morel-Nagaraaj type mixture
  K  = number of variates required, integer > 0
  m  = index, integer > 0,   scalar or vector
  pi = parameter, in (0,1),  scalar or vector
  rho= parameter, in (0,1),  scalar or vector
  If any of m, pi and rho is a vector, it must be column vector, K by 1
*/

  alpha = sqrt(rho)#(1-pi) + pi;
  beta  = pi#(1-sqrt(rho));
  seed = j(K, 1, &seed);
  z = (ranuni(seed) <= pi);
  y = z     # ranbin(seed, m, alpha) +
      (1-z) # ranbin(seed, m, beta);
  return (y);
finish;

********************************************************************;
start beta_v (K, pi, rho);
/*
  Returns a col vector of K random variates from the beta distrib
  K  = number of variates required, integer > 0
  pi = parameter, in (0,1),  scalar or vector
  rho= parameter, in [0,1),  scalar or vector
  If pi or  rho is a vector, it must be column vector, K by 1
*/

  if (rho = 0) then return(j(K, 1, pi));

  c = (1-rho) / rho;
  a = pi # c;
  b = (1 - pi) # c;
  if (nrow(a) = 1) then
     do; a = j(K, 1, a); b = j(K, 1, b); end;
  p = rand("beta", a, b);
  return (p);
finish;

********************************************************************;
start madsen_1(m, pi, rho);
/*
  returns a random variate from the Madsen-type mixture
  m  = index (integer scalar)
  pi = parameter (scalar in (0,1))
  rho= parameter (scalar in (0,1))
*/
  if (ranuni(&seed) <= rho) then y = m*(ranuni(&seed) <= pi);
  else y = ranbin(&seed, m, pi);
  return (y);
finish;

********************************************************************;
start madsen_v(K, m, pi, rho);
/*
  Returns a col vector of K random variates from the Madsen-type mixture
  K  = number of variates required, integer > 0
  m  = index, integer > 0,   scalar or vector
  pi = parameter, in (0,1),  scalar or vector
  rho= parameter, in (0,1),  scalar or vector
  If any of m, pi and rho is a vector, it must be column vector, K by 1
*/

  seed = j(K, 1, &seed);
  z = (ranuni(seed) <= rho);
  y = z # m # (ranuni(seed) <= pi) + (1 - z) # ranbin(seed, m, pi);
  return (y);
finish;

********************************************************************;
start sim_ebv_v (method, K, m, pi, rho);
/*
  Returns a col vector of K random variates from the specifed method
  method = method name (string), see allowed names in the code.
  K  = number of variates required, integer > 0
  m  = index, integer > 0,   scalar or vector
  pi = parameter, in (0,1),  scalar or vector
  rho= parameter, in (0,1),  scalar or vector
  If any of m, pi and rho is a vector, it must be column vector, K by 1
*/

  if      (upcase(method) = "MADSEN") then
    y = madsen_v(K, m, pi, rho);
  else if (upcase(method) = "MORELNAG") then
    y = morelnag_v(K, m, pi, rho);
  else if (upcase(method) = "BETABIN") then
    y = betabin_v(K, m, pi, rho);
  else if (upcase(method) = "BINOMIAL") then
    y = binomial_v(K, m, pi, rho);
  else
    y = .;   /* error: unrecognized name  */

  return (y);

finish;

********************************************************************;
start   tnegbin(K, x, p, trunc);
/*
  returns K observations, col vec, from a negbin with
  values <= "trunc" are excluded (truncated out).
*/
  if (&sysver >= 9) then p_ = p; else p_ = 1 - p;
      * use "q" in SAS 8, "p" in SAS 9;

  do until (nrow(m) >= K);
    do until (nrow(y) > 0);
      y = rand("negb", j(K*1.5, 1, x), p_);
      l = loc(y > trunc);
      if (ncol(l) > 0) then y = y[l];
    end;
    m = m // y;
  end;

  return (m[1:K]);
finish;

********************************************************************;
start Mardia(a, b, c);
/*
A specialized quadratic solver.
Solves (a x^2 + b x + c = 0). Assumes a is NOT 0.
returns: <- the root with "-" (Mardia's formula)
*/

* if (a=0) then return(-c/b);   *  we assume a is NOT 0;

  if (b>0) then s=1; else if (b<0) then s=-1; else s=0;
  p  = -0.5 * (b + s*sqrt(b*b-4*a*c));
  r1 = p/a;
  r2 = c/p;

  if (r2>0) then return (r2); else return (r1);

finish;
************************************************************;
start solv2 (ui, uj, psi);
/*
input: ui = E[Y_i]
input: uj = E[Y_j]
input: psi= O.R.(Y_i, Y_j)
returns: <- E[Y_i Y_j]
*/
  if (psi=1) then uij = ui*uj;
  else do;
    a   = 1 - psi;
    b   = 1 - a * (ui + uj);
    c   = -psi * ui * uj;
    uij = Mardia(a, b, c);  * Mardia's root;
  end;

  return (uij);

finish;
